from torchmetrics import Metric
import torch
from typing import Tuple, Set, List, Optional

class Triple(tuple):
    """存放三元组的五元形式(sub_start, sub_end, predicate, obj_start, obj_end), 这样改写方便求交集"""
    
    def __init__(self, triple: Tuple[int, str]):
        super().__init__()
        self.triple = triple
        
    def __hash__(self):
        return self.triple.__hash__()

    def __eq__(self, triple):
        return self.triple == triple.triple

    def __repr__(self) -> str:
        return str(self.triple)

    def __len__(self) -> int:
        return len(self.triple)

    def __getitem__(self, index):
        return self.triple[index]

    @property
    def subject(self):
        """返回subject的下标, 左闭右开"""
        return (self.triple[0], self.triple[1])
       
    @property
    def predicate(self):
        return self.triple[2]
        
    
    @property
    def object(self):
        """返回object的下标, 左闭右开"""
        return (self.triple[3], self.triple[4])

class TripleF1(Metric):
    """三元组抽取的F1
    说明:
    - 输入为(subject_start, subject_end, predicate, object_start, object_end)
    """

    full_state_update: Optional[bool] = False

    def __init__(self):
        super().__init__()
        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: List[Set[Triple]], true: List[Set[Triple]]) -> None:
        """
        参数:
        - pred: batch_size大小的列表, 列表内为每个样本的预测的三元组的集合(set)
        - true: 同pred
        """
        assert len(pred) == len(true), f'pred : {pred}, true : {true}'
        for p, t in zip(pred, true):
            self.correct += len(p & t)
            self.all_pred += len(p)
            self.all_true += len(t)
    
    def compute(self):
        # 下面这个公式可以通过化简得到
        return 2 * self.correct / (self.all_pred + self.all_true + 1e-10)


from .text_classification import TextClassificationDataModule
from .text_pair_classification import TextPairClassificationDataModule
from .text_pair_regression import TextPairRegressionDataModule
from .token_classification import TokenClassificationDataModule
from .span_extraction import SpanExtractionDataModule
from .relation_extraction import RelationExtractionDataModule
from .prompt_span_extraction import PromptSpanExtractionDataModule
from .prompt_relation_extraction import PromptRelationExtractionDataModule
from .event_extraction import EventExtractionDataModule
from .entity_extraction import EntityExtractionDataModule
from .question_answering import QuestionAnsweringDataModule



from ..utils.make_datamodule import PLMBaseDataModule, sequence_padding
import torch


class QuestionAnsweringDataModule(PLMBaseDataModule):
    """span分类的数据模块 数据集必须有text, spans两个字段
    """
    def __init__(self,
                dataset: str,
                batch_size: int,
                plm: str = 'hfl/chinese-roberta-wwm-ext',
                **kwargs):
        super().__init__()
        
    def gp_transform(self, examples):
        batch_text = examples['text']
        batch_question = examples['question']
        batch_spans = examples['spans']
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_span_ids = []
        batch_inputs = self.tokenizer(batch_question,
                                      batch_text, 
                                      padding=True,  
                                      truncation=True,
                                      max_length=max_length,
                                      return_tensors='pt')
        for i, text in enumerate(batch_text):
            spans = batch_spans[i]
            span_ids = torch.zeros(1, max_length, max_length)
            for span in spans :
                start = span['indices'][0]
                _start = batch_inputs.char_to_token(i, start, 1)
                end = span['indices'][-1] 
                _end = batch_inputs.char_to_token(i, end, 1)
                span_ids[0, _start, _end] = 1
            batch_span_ids.append(span_ids)
        batch_span_ids = torch.stack(batch_span_ids, dim=0)
        batch_inputs['span_tags'] = batch_span_ids
        return batch_inputs
    
    def sequence_transform(self, examples):
        batch_text = examples['text']
        batch_question = examples['question']
        batch_spans = examples['spans']
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_tags = []
        batch_inputs = self.tokenizer(batch_question,
                                      batch_text, 
                                      padding=True,  
                                      truncation=True,
                                      max_length=max_length,
                                      return_tensors='pt')
        for i, text in enumerate(batch_text):
            spans = batch_spans[i]
            tags = torch.zeros(max_length, dtype=torch.long)
            for span in spans :
                for idx in span['indices']: 
                    token_idx = batch_inputs.char_to_token(i, idx, 1)
                    tags[token_idx] = 1
            batch_tags.append(tags)
        batch_tags = torch.stack(batch_tags)
        batch_inputs['tags'] = batch_tags
        return batch_inputs
    
    def pointer_transform(self, examples):
        batch_text = examples['text']
        batch_question = examples['question']
        batch_spans = examples['spans']
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_start_tags = []
        batch_end_tags = []
        batch_inputs = self.tokenizer(batch_question,
                                      batch_text, 
                                      padding=True,  
                                      truncation=True,
                                      max_length=max_length,
                                      return_tensors='pt')
        for i, text in enumerate(batch_text):
            spans = batch_spans[i]
            start_tags = torch.zeros(max_length, dtype=torch.long)
            end_tags = torch.zeros(max_length, dtype=torch.long)
            for span in spans :
                for idx in span['indices'][0:]:
                    _start = batch_inputs.char_to_token(batch_or_char_index=i, char_index=idx, sequence_index=1)
                    if _start is not None:
                        break
                for idx in span['indices'][::-1]:
                    _end = batch_inputs.char_to_token(batch_or_char_index=i, char_index=idx, sequence_index=1)
                    if _end is not None:
                        break
                start_tags[_start] = 1
                end_tags[_end] = 1
            batch_start_tags.append(start_tags)
            batch_end_tags.append(end_tags)
        batch_start_tags = torch.stack(batch_start_tags)
        batch_end_tags = torch.stack(batch_end_tags)
        batch_inputs['start_tags'] = batch_start_tags
        batch_inputs['end_tags'] = batch_end_tags
        return batch_inputs


import torch
from torchmetrics import Metric
from typing import Optional

def end_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk ended between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_end: boolean.
    """
    chunk_end = False

    if prev_tag == 'S':
        chunk_end = True
    # pred_label中可能出现这种情形
    if prev_tag == 'B' and tag == 'B':
        chunk_end = True
    if prev_tag == 'B' and tag == 'S':
        chunk_end = True
    if prev_tag == 'B' and tag == 'O':
        chunk_end = True
    if prev_tag == 'I' and tag == 'B':
        chunk_end = True
    if prev_tag == 'I' and tag == 'S':
        chunk_end = True
    if prev_tag == 'I' and tag == 'O':
        chunk_end = True

    if prev_tag != 'O' and prev_tag != '.' and prev_type != type_:
        chunk_end = True

    return chunk_end


def start_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk started between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_start: boolean.
    """
    chunk_start = False

    if tag == 'B':
        chunk_start = True
    if tag == 'S':
        chunk_start = True

    if prev_tag == 'S' and tag == 'I':
        chunk_start = True
    if prev_tag == 'O' and tag == 'I':
        chunk_start = True

    if tag != 'O' and tag != '.' and prev_type != type_:
        chunk_start = True

    return chunk_start

def get_entities(seq):
    """
    Gets entities from sequence.
    Args:
        seq (list): sequence of labels.
    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).
    Example:
        seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        get_entities(seq)
        [('PER', 0, 1), ('LOC', 3, 3)]
    """
    # for nested list
    if any(isinstance(s, list) for s in seq):
        seq = [item for sublist in seq for item in sublist + ['O']]
    prev_tag = 'O'
    prev_type = ''
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq + ['O']):
        tag = chunk[0]
        type_ = chunk.split('-')[-1]

        if end_of_chunk(prev_tag, tag, prev_type, type_):
            chunks.append((prev_type, begin_offset, i - 1))
        if start_of_chunk(prev_tag, tag, prev_type, type_):
            begin_offset = i
        prev_tag = tag
        prev_type = type_

    return chunks



class ChunkF1(Metric):

    full_state_update: Optional[bool] = False
    
    def __init__(self):
        super().__init__()
        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred, true):
        true_entities = set(get_entities(true))
        pred_entities = set(get_entities(pred))
        self.correct += len(true_entities & pred_entities)
        self.all_pred += len(pred_entities)
        self.all_true += len(true_entities)


    def compute(self):
        return 2 * self.correct / (self.all_pred + self.all_true)


import torch.nn as nn
import torch

class LayerNorm(nn.Module):
    def __init__(self, 
                 hidden_size:int, 
                 conditional_size:int,
                 eps:float=1e-12, 
                 weight:bool =True,
                 bias:bool = True,
                 norm_mode:str = 'normal',
                 **kwargs):
        """layernorm 层，这里自行实现，目的是为了兼容 conditianal layernorm，使得可以做条件文本生成、条件分类等任务
           条件layernorm来自于苏剑林的想法，详情：https://spaces.ac.cn/archives/7124
        """
        super(LayerNorm, self).__init__()
        if weight:
            self.weight = nn.Parameter(torch.ones(hidden_size))
        if bias:
            self.bias = nn.Parameter(torch.zeros(hidden_size))
        self.norm_mode = norm_mode

        self.eps = eps
        self.conditional_size = conditional_size
        if conditional_size:
            # 条件layernorm, 用于条件文本生成,
            # 这里采用全零初始化, 目的是在初始状态不干扰原来的预训练权重
            self.dense1 = nn.Linear(conditional_size, hidden_size, bias=False)
            self.dense1.weight.data.uniform_(0, 0)
            self.dense2 = nn.Linear(conditional_size, hidden_size, bias=False)
            self.dense2.weight.data.uniform_(0, 0)

    def forward(self, inputs, cond=None):
        """
        Args:
            inputs (B, L, H): 特征输入.
            cond (_type_, optional): 条件输入. Defaults to None.
        """
        if self.norm_mode == 'rmsnorm':
            # t5使用的是RMSnorm
            variance = inputs.to(torch.float32).pow(2).mean(-1, keepdim=True)
            o = inputs * torch.rsqrt(variance + self.eps)
        else:
            # 归一化是针对于inputs
            u = inputs.mean(-1, keepdim=True)
            s = (inputs - u).pow(2).mean(-1, keepdim=True)
            o = (inputs - u) / torch.sqrt(s + self.eps)

        if not hasattr(self, 'weight'):
            self.weight = 1
        if not hasattr(self, 'bias'):
            self.bias = 0

        if self.conditional_size:
            # 三者的形状都是一致的
            for _ in range(len(inputs.shape) - len(cond.shape)):
                cond = cond.unsqueeze(dim=1)
            return (self.weight + self.dense1(cond)) * o + (self.bias + self.dense2(cond))
        else:
            return self.weight * o + self.bias


import pytorch_lightning as pl
from transformers import BertModel, BertTokenizer
from ...layers import SimpleDense
import torch.nn as nn
from torchmetrics import F1Score
import torch


class BertTextMultiClassification(pl.LightningModule):
    """基于bert的多标签文本分类"""
    def __init__(
        self,
        hidden_size: int,
        lr: float,
        dropout: float,
        weight_decay: float,
        threshold: float = 0.5,
        **data_params
    ):
        super(BertTextMultiClassification, self).__init__()
        self.save_hyperparameters()
        self.bert = BertModel.from_pretrained(data_params['pretrained_dir'] + data_params['pretrained_model'])
        self.dropout = nn.Dropout(p=dropout)
        self.classifier = SimpleDense(self.bert.config.hidden_size, hidden_size, len(data_params['label2id']))

        self.criterion = nn.BCEWithLogitsLoss()
        self.train_f1 = F1Score(num_classes=len(data_params['label2id']), average='macro', multiclass=True, mdmc_average='global')
        self.val_f1 = F1Score(num_classes=len(data_params['label2id']), average='macro', multiclass=True, mdmc_average='global')
        self.test_f1 = F1Score(num_classes=len(data_params['label2id']), average='macro', multiclass=True, mdmc_average='global')

        self.tokenizer = BertTokenizer.from_pretrained(data_params['pretrained_dir'] + data_params['pretrained_model'])

    def forward(self, input_ids, token_type_ids, attention_mask):
        pooled = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).pooler_output
        pooled = self.dropout(pooled)
        logits = self.classifier(pooled)
        return logits


    def shared_step(self, batch):
        inputs, label_ids = batch['inputs'], batch['label_ids']
        logits = self(**inputs)
        loss = self.criterion(logits, label_ids)
        pred_ids = logits.sigmoid().ge(self.hparams.threshold).long()
        return loss,  pred_ids, label_ids

    def training_step(self, batch, batch_id):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.train_f1(pred_ids, label_ids.long())
        self.log('train/f1', self.train_f1, on_step=True, on_epoch=True,  prog_bar=True)
        return {'loss': loss}


    def validation_step(self, batch, batch_id):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.val_f1(pred_ids, label_ids.long())
        self.log('val/f1', self.val_f1, on_step=False, on_epoch=True,  prog_bar=True)
        return {'val_loss': loss}

    def test_step(self, batch, batch_id):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.test_f1(pred_ids, label_ids.long())
        self.log('test/f1', self.test_f1, on_step=False, on_epoch=True,  prog_bar=True)
        return {'test_loss': loss}


    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(self.parameters(), lr=self.hparams.lr, weight_decay=self.hparams.weight_decay)
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [optimizer], [scheduler]



import torch.nn as nn
import torch


class OneRelSpanClassifier(nn.Module):
    """枚举出sel_len * sel_len个span,并进行分类

    Args:
        input_size(int): 输入特征维度
        output_size(int): 分类的类别数量
    输出形状: [batch, seq, seq, output_size]
    """
    
    def __init__(self,
                 input_size: int,
                 output_size: int,
                 ):
        super().__init__()
        self.output_size = output_size
        self.linear1 = nn.Linear(input_size*2, input_size)
        self.activation = nn.ReLU()
        self.linear2 = nn.Linear(input_size, output_size)
        self.layernorm = nn.LayerNorm(input_size)
        
    def forward(self, x): 
        batch_size, seq_len, hidden_size = x.size()
        # head: [batch_size, seq_len * seq_len, hidden_size] 重复样式为1,1,1, 2,2,2, 3,3,3
        head = x.unsqueeze(2).expand(batch_size, seq_len, seq_len, hidden_size).reshape(batch_size, seq_len*seq_len, hidden_size)
        # tail: [batch_size, seq_len * seq_len, hidden_size] 重复样式为1,2,3, 1,2,3, 1,2,3
        tail = x.repeat(1, seq_len, 1)
        # 两两token的特征拼接在一起
        pairs = torch.cat([head, tail], dim=-1)
        pairs = self.linear1(pairs)
        pairs = self.layernorm(pairs)
        pairs = self.activation(pairs)
        scores = self.linear2(pairs).reshape(batch_size, seq_len, seq_len, self.output_size)
        return scores


from typing import List, Optional
from omegaconf import DictConfig
from .utils import utils
from lightning.pytorch import (Callback,
                               LightningDataModule,
                               LightningModule,
                               Trainer,
                               seed_everything)
from lightning.pytorch.loggers import Logger
import hydra
import os

os.environ["TOKENIZERS_PARALLELISM"] = "false"

log = utils.get_logger(__name__)


@hydra.main(config_path="configs/", config_name="config.yaml", version_base='1.1')
def run(config: DictConfig) -> Optional[float]:
    """Contains training pipeline.
    Instantiates all PyTorch Lightning objects from config.

    Args:
        config (DictConfig): Configuration composed by Hydra.

    Returns:
        Optional[float]: Metric score for hyperparameter optimization.
    """
    
    utils.extras(config)
    if config.get("print_config"):
        utils.print_config(config, resolve=True)
        
    if config.get("use_hf_mirror"):
        os.environ["HF_ENDPOINT"] = "https://hf-mirror.com"
        log.info("Using hf-mirror to download datasets and models!")
        
    # Set seed for random number generators in pytorch, numpy and python.random
    if config.get("seed"):
        seed_everything(config.seed, workers=True)

    # Init lightning datamodule
    if len(config.datamodule) > 0:
        log.info(f"Instantiating datamodule <{config.datamodule._target_}>")
        datamodule: LightningDataModule = hydra.utils.instantiate(config.datamodule)
        datamodule.prepare_data()
        datamodule.setup(stage='fit')
    else : datamodule = None

    # Init lightning model
    if datamodule is not None:
        log.info(f"Instantiating model <{config.model._target_}>")
        model: LightningModule = hydra.utils.instantiate(config.model, **datamodule.hparams)
    else: 
        log.info(f"Instantiating model <{config.model._target_}>")
        model = hydra.utils.instantiate(config.model)

    # Init lightning callbacks
    callbacks: List[Callback] = []
    if "callbacks" in config:
        for _, cb_conf in config.callbacks.items():
            if "_target_" in cb_conf:
                log.info(f"Instantiating callback <{cb_conf._target_}>")
                callbacks.append(hydra.utils.instantiate(cb_conf))

    # Init lightning loggers
    logger: List[Logger] = []
    if "logger" in config:
            log.info(f"Instantiating logger <{config.logger._target_}>")
            logger.append(hydra.utils.instantiate(config.logger))

    # Init lightning trainer
    log.info(f"Instantiating trainer <{config.trainer._target_}>")
    trainer: Trainer = hydra.utils.instantiate(config.trainer, 
                                               callbacks=callbacks, 
                                               logger=logger)

    # Send some parameters from config to all lightning loggers
    log.info("Logging hyperparameters!")
    utils.log_hyperparameters(config=config,
                              model=model,
                              datamodule=datamodule,
                              trainer=trainer,
                              callbacks=callbacks,
                              logger=logger)


    # Train the model
    log.info("Starting training!")
    trainer.fit(model=model, datamodule=datamodule)

    # Get metric score for hyperparameter optimization
    optimized_metric = config.get("optimized_metric")
    if optimized_metric and optimized_metric not in trainer.callback_metrics:
        raise Exception(
            "Metric for hyperparameter optimization not found! "
            "Make sure the `optimized_metric` in `hparams_search` config is correct!"
        )
    score = trainer.callback_metrics.get(optimized_metric)

    # Test the model
    if config.get("test_after_training") and not config.trainer.get("fast_dev_run"):
        log.info("Starting testing!")
        datamodule.setup(stage='test')
        trainer.test(model=model, datamodule=datamodule, ckpt_path="best")

    # Make sure everything closed properly
    log.info("Finalizing!")
    utils.finish(config=config,
                 model=model,
                 datamodule=datamodule,
                 trainer=trainer,
                 callbacks=callbacks,
                 logger=logger)

    # Print path to best checkpoint
    if not config.trainer.get("fast_dev_run"):
        log.info(f"Best model ckpt at {trainer.checkpoint_callback.best_model_path}")

    # Return metric score for hyperparameter optimization
    return score


if __name__ == "__main__":
    run()


from functools import lru_cache
from typing import Tuple, List, Dict
from ..utils.make_datamodule import PLMBaseDataModule
from ..utils import utils
import torch

log = utils.get_logger(__name__)


class TextClassificationDataModule(PLMBaseDataModule):
    """单文本分类数据模块
    
    Args:
        dataset (str): 数据集名称
        plm (str): 预训练模型名称
        batch_size (int): 批次大小
        transform (str): 数据转换形式,默认为simple
        num_workers (int, optional): 进程数. Defaults to 0.
        pin_memory (bool, optional): . Defaults to False.
        plm_dir (str, optional): 自定义预训练模型文件夹. Defaults to './plms/'.
        dataset_dir (str, optional): 自定义数据集文件夹. Defaults to './datasets/'.
        
    """
    def __init__(self,
                dataset: str,
                batch_size: int ,
                plm: str = 'hfl/chinese-roberta-wwm-ext',
                **kwargs):
        super().__init__()       
        
    def setup(self, stage: str) -> None:
        self.hparams.id2label = self.id2label

    
    def bert_transform(self, examples) -> Dict:
        batch_text = examples['text']
        batch_label_ids = []
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_inputs = self.tokenizer(batch_text,
                                      padding=True,
                                      max_length=max_length,
                                      truncation=True,
                                      return_tensors='pt')
        for i, text in enumerate(batch_text):
            label_id = self.label2id[examples['label'][i]]
            batch_label_ids.append(label_id)
        batch_inputs['label_ids'] = torch.LongTensor(batch_label_ids)        
        return batch_inputs
    
    
    @property
    @lru_cache()
    def labels(self):
        return sorted(set(self.train_df.label.values))


    @property
    def label2id(self):
        label2id = {l : i for i,l in enumerate(self.labels)}
        return label2id


    @property
    def id2label(self):
        return {i:l for l,i in enumerate(self.label2id)}

        
    @classmethod
    def get_one_example(cls):
        return {'label': '新闻', 'text': '怎么给这个图片添加超级链接呢？'}


from ...utils.make_model import PLMBaseModel
from ...data import Entity
from ...layers.classifier.biaffine import EfficientBiaffineSpanClassifier
from ...layers.dropout import MultiDropout
from ...layers.loss import MultiLabelCategoricalCrossEntropy
from ...metrics.span import SpanF1
import torch
from typing import List


class BiaffineForEntityExtraction(PLMBaseModel):
    """span级别的实体识别,支持嵌套实体识别

    Args:
        lr: 学习率. 默认3e-5
        scheduler: 学习率衰减器. 
        hidden_size: 中间层维度. 默认128
        threshold: 抽取阈值.抽取大于这个值的span
        weight_decay: 权重衰减
    """
    def __init__(self,
                 lr: float = 3e-5,
                 hidden_size: int = 64,
                 threshold: float = 0.0,
                 scheduler: str = 'linear_warmup',
                 weight_decay: float = 0.01,
                 **kwargs) -> None:
        super().__init__()
        
        self.plm = self.get_plm_architecture(add_pooler_layer=False)
        self.classifier = EfficientBiaffineSpanClassifier(input_size=self.plm.config.hidden_size,
                                                            hidden_size=self.hparams.hidden_size,
                                                            output_size=len(self.hparams.id2ent))
        self.criterion = MultiLabelCategoricalCrossEntropy()

        self.train_metric = SpanF1()
        self.val_metric = SpanF1()
        self.test_metric = SpanF1()
        
        
    def setup(self, stage: str) -> None:
        self.trainer.datamodule.dataset.set_transform(self.trainer.datamodule.tp_transform)


    def forward(self, input_ids, attention_mask):
        x = self.plm(input_ids=input_ids, attention_mask=attention_mask).last_hidden_state
        x = self.classifier(x, mask=attention_mask)
        return x
    
    
    def step(self, batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']
        span_ids = batch['tag_ids']
        logits = self(input_ids, attention_mask)
        loss = self.criterion(logits.reshape(logits.shape[0]*logits.shape[1], -1), span_ids.reshape(span_ids.shape[0]*span_ids.shape[1], -1))
        pred = logits.ge(self.hparams.threshold).float()
        return loss, pred, span_ids

        
    def training_step(self, batch, batch_idx):
        loss, pred, true = self.step(batch)
        self.train_metric(pred, true)
        self.log('train/f1', self.train_metric, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss':loss}
    
    
    def validation_step(self, batch, batch_idx):
        loss, pred, true = self.step(batch)
        self.val_metric(pred, true)
        self.log('val/f1', self.val_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss':loss}
    
    
    def test_step(self, batch, batch_idx):
        loss, pred, true = self.step(batch)
        self.test_metric(pred, true)
        self.log('test/f1', self.test_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}
    
    
    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler_config = self.get_scheduler_config(optimizer, self.hparams.scheduler)
        return [optimizer], [scheduler_config]
    
    
    def predict(self, text: str, device: str='cpu') -> List[Entity]:
        threshold = self.hparams.threshold
        inputs = self.tokenizer(text,
                                max_length=self.hparams.plm_max_length,
                                truncation=True,
                                return_token_type_ids=False,
                                return_tensors='pt')
        inputs.to(device)
        preds = self(**inputs)
        spans_ls = torch.nonzero(preds>threshold).tolist()
        ents = []
        for span in spans_ls :
            start_token = span[2]
            end_token = span[3]
            start_offset = inputs.token_to_chars(start_token)
            end_offset = inputs.token_to_chars(end_token)
            start = start_offset[0]
            end = end_offset[-1]
            label = self.hparams.id2ent[span[1]]
            ents.append(Entity(label=label, indices=[i for i in range(start, end)], text=text[start: end]))
        return ents


from .normalization import LayerNorm
from .loss import MultiLabelCategoricalCrossEntropy, SparseMultiLabelCrossEntropy
from .classifier import GlobalPointer, EfficientGlobalPointer, CRF, SimpleDense, Biaffine, BiaffineSpanClassifier, EfficientBiaffineSpanClassifier
from .embedding import SinusoidalPositionEmbedding
from .activation import GELU, SWISH, GELU_Approximate
from .bert import Bert, BertEmbeddings, BertAttention, BertAddNorm, BertEncoder, BertPooler
from .word2vec import SkipGram
from .dropout import MultiDropout






from ...layers import EfficientGlobalPointer, MultiDropout
from ...layers.loss import MultiLabelCategoricalCrossEntropy
from ...metrics.triple import TripleF1, Triple
from ...utils.make_model import align_token_span, PLMBaseModel
import torch
from torch import Tensor
from typing import List, Set


class GPLinkerForPromptRelationExtraction(PLMBaseModel):
    """基于globalpointer的关系抽取模型
    参考:
    - https://kexue.fm/archives/8888
    - https://github.com/bojone/bert4keras/blob/master/examples/task_relation_extraction_gplinker.py
    """
    def __init__(
        self,
        hidden_size: int,
        lr: float,
        dropout: float,
        weight_decay: float,
        threshold: float,
        scheduler: str = 'linear_warmup_step',
        **kwargs):
        super().__init__()        
        
        self.plm = self.get_plm_architecture()
        self.dropout = MultiDropout()
        
        # 主语 宾语分类器
        self.so_classifier = EfficientGlobalPointer(self.plm.config.hidden_size, hidden_size, 2)
        # 主语 宾语 头对齐
        self.head_classifier = EfficientGlobalPointer(self.plm.config.hidden_size, hidden_size, 1, RoPE=False, tril_mask=False)
        # 主语 宾语 尾对齐
        self.tail_classifier = EfficientGlobalPointer(self.plm.config.hidden_size, hidden_size, 1, RoPE=False, tril_mask=False)

        self.so_criterion = MultiLabelCategoricalCrossEntropy()
        self.head_criterion = MultiLabelCategoricalCrossEntropy()
        self.tail_criterion = MultiLabelCategoricalCrossEntropy()

        self.train_metric = TripleF1()
        self.val_metric = TripleF1()
        self.test_metric = TripleF1()

    

    def forward(self, input_ids, token_type_ids=None, attention_mask=None):
        hidden_state = self.plm(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        hidden_state = self.dropout(hidden_state)
        so_logits = self.so_classifier(hidden_state, mask=attention_mask)
        head_logits = self.head_classifier(hidden_state, mask=attention_mask)
        tail_logits = self.tail_classifier(hidden_state, mask=attention_mask)
        return so_logits, head_logits, tail_logits


    def shared_step(self, batch):
        #inputs为bert常规输入, span_ids: [batch_size, 2, seq_len, seq_len],
        input_ids, token_type_ids, attention_mask = batch['input_ids'], batch['token_type_ids'], batch['attention_mask']
        so_ture, head_true, tail_true = batch['so_ids'],  batch['head_ids'], batch['tail_ids']
        so_logits, head_logits, tail_logits = self(input_ids, token_type_ids, attention_mask)


        so_logits_ = so_logits.reshape(so_logits.shape[0] * so_logits.shape[1], -1)
        so_true_ = so_ture.reshape(so_ture.shape[0] * so_ture.shape[1], -1)
        so_loss = self.so_criterion(so_logits_, so_true_)
        
        head_logits_ = head_logits.reshape(head_logits.shape[0] * head_logits.shape[1], -1)
        head_true_ = head_true.reshape(head_true.shape[0] * head_true.shape[1], -1)
        head_loss = self.head_criterion(head_logits_, head_true_)
        
        tail_logits_ = tail_logits.reshape(tail_logits.shape[0] * tail_logits.shape[1], -1)
        tail_true_ = tail_true.reshape(tail_true.shape[0] * tail_true.shape[1], -1)
        tail_loss = self.tail_criterion(tail_logits_, tail_true_)
        
        loss = so_loss + head_loss + tail_loss
        return loss, so_logits, head_logits, tail_logits

        
    def training_step(self, batch, batch_idx) -> dict:
        # 训练阶段不进行解码, 会比较慢
        loss, so_logits, head_logits, tail_logits = self.shared_step(batch)
        self.log('train/loss', loss)
        return {'loss': loss}


    def validation_step(self, batch, batch_idx) -> dict:
        so_true, head_true, tail_true = batch['so_ids'], batch['head_ids'], batch['tail_ids']
        loss, so_logits, head_logits, tail_logits = self.shared_step(batch)
        batch_triples = self.extract_triple(so_logits, head_logits, tail_logits, threshold=self.hparams.threshold)
        batch_true_triples = self.extract_triple(so_true, head_true, tail_true, threshold=self.hparams.threshold)
        self.val_metric(batch_triples, batch_true_triples)
        self.log('val/f1', self.val_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'val_loss': loss}


    def test_step(self, batch, batch_idx) -> dict:
        so_true, head_true, tail_true = batch['so_ids'], batch['head_ids'], batch['tail_ids']
        loss, so_logits, head_logits, tail_logits = self.shared_step(batch)
        batch_triples = self.extract_triple(so_logits, head_logits, tail_logits, threshold=self.hparams.threshold)
        batch_true_triples = self.extract_triple(so_true, head_true, tail_true, threshold=self.hparams.threshold)
        self.test_metric(batch_triples, batch_true_triples)
        self.log('test/f1', self.test_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'test_loss': loss}


    def configure_optimizers(self)  :
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.so_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.so_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0},
            {'params': [p for n, p in self.head_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.head_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0},
            {'params': [p for n, p in self.tail_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.tail_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters, eps=1e-5)
        scheduler_config = self.get_scheduler_config(optimizer=optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler_config]


    def extract_triple(
        self, 
        so_logits: Tensor,
        head_logits: Tensor, 
        tail_logtis: Tensor, 
        threshold: float
        ) -> List[Set[Triple]]:
        """
        将三个globalpointer预测的结果进行合并，得到三元组的预测结果
        参数:
        - so_logits: [batch_size, 2, seq_len, seq_len]
        - head_logits: [batch_size, 1, seq_len, seq_len]
        - tail_logtis: [batch_size, 1, seq_len, seq_len]
        返回:
        - batch_size大小的列表，每个元素是一个集合，集合中的元素是三元组
        """

        so_logits = so_logits.chunk(so_logits.shape[0])
        head_logits = head_logits.chunk(head_logits.shape[0])
        tail_logtis = tail_logtis.chunk(tail_logtis.shape[0])
        assert len(head_logits) == len(tail_logtis) == len(so_logits)
        batch_triples = []
        for i in range(len(so_logits)):
            subjects, objects = set(), set()
            for l, h, t in zip(*torch.where(so_logits[i].squeeze(0) > threshold)):
                if  l == 0:
                    subjects.add((h, t))
                else:
                    objects.add((h, t))
            
            triples = set()
            for sh, st in subjects:
                for oh, ot in objects:
                    p1s = torch.where(head_logits[i].squeeze(0)[:, sh, oh] > threshold)[0].tolist()
                    p2s = torch.where(tail_logtis[i].squeeze(0)[:, st, ot] > threshold)[0].tolist()
                    ps = set(p1s) & set(p2s)
                    if len(ps) > 0:
                        for p in ps:
                            triples.add(Triple(triple=(sh.item(), st.item(), '', oh.item(), ot.item())))
            batch_triples.append(triples)
        return batch_triples




    def predict(self,prompts:List[str], texts: List[str], device:str='cpu', threshold: float = None) -> Set[Triple]:
        """模型预测
        参数:
        - text: 要预测的单条文本
        - device: 设备
        - threshold: 三元组抽取阈值, 如果为None, 则为模型训练时的阈值
        返回
        - 预测的三元组
        """
        max_length = min(max([len(prompt+text)+3 for prompt, text in zip(prompts, texts)]), 512)
        inputs = self.tokenizer(
            prompts,
            texts,
            max_length=max_length,
            truncation=True,
            return_tensors='pt',
            padding='max_length')
        mapping = self.tokenizer(
            prompts,
            texts,
            max_length=max_length,
            truncation=True,
            padding='max_length',
            return_offsets_mapping=True)['offset_mapping']
        inputs.to(torch.device(device))
        so_logits, head_logits, tail_logits = self(**inputs)
        if threshold is None:
            threshold = self.hparams.threshold
        batch_triples = self.extract_triple(so_logits, head_logits, tail_logits, threshold=threshold)
        rels = []
        if len(batch_triples) >0:
            for i, triples_type in enumerate(batch_triples):
                triples = [(triple[0], triple[1], prompts[i], triple[3], triple[4])  for triple in triples_type]
                for triple in triples:
                    sub = align_token_span((triple[0], triple[1]+1), mapping[i])
                    obj = align_token_span((triple[3], triple[4]+1), mapping[i])
                    rels.append((sub[0],sub[1],prompts[i],obj[0],obj[1]))
        return rels


from torch import Tensor
from torchmetrics import Metric
import torch
from typing import List, Set, Optional


class SpanF1(Metric):
    """计算span矩阵的F1"""

    full_state_update: Optional[bool] = False

    def __init__(self):
        super().__init__()

        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: Tensor, true: Tensor):
        self.correct += torch.sum(pred[true==1])
        self.all_pred += torch.sum(pred == 1)
        self.all_true += torch.sum(true == 1)

    def compute(self):
        return 2 * self.correct / (self.all_pred + self.all_true)
    
    
class SpanTokenF1(Metric):
    """以预测的span中的token正确为评判指标
    - 输入的形式为[0,0,0,1,1,1]
    """
    
    full_state_update: Optional[bool] = False
    
    def __init__(self):
        super().__init__()

        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: Tensor, true: Tensor):
        self.correct += torch.sum(pred[true==1])
        self.all_pred += torch.sum(pred == 1)
        self.all_true += torch.sum(true == 1)

    def compute(self):
        return 2 * self.correct / (self.all_pred + self.all_true)
    
    
class SpanIndexF1(Metric):
    """以预测的下标正确为评判指标
    - 输入的形式为[{0,1,2}, {0,1,2}]
    """
    
    full_state_update: Optional[bool] = False
    
    def __init__(self):
        super().__init__()

        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: List[Set[int]], true: List[Set[int]]):
        for p, t in zip(pred, true):
            if len(t) == 0 and len(p) == 0: # 如果标签和预测的下标都为0,则测试F1为1
                self.correct += 1
                self.all_pred += 1
                self.all_true += 1
            else:
                self.correct += len(p & t)
                self.all_pred += len(p)
                self.all_true += len(t)

    def compute(self):
        return 2 * self.correct / (self.all_pred + self.all_true)


import torch 
import torch.nn as nn 
import torch.nn.functional as F

# class CRF(nn.Module):
#     """
#     CRFlayer
#     """
#     def __init__(self, output_dim):
#         super(CRF, self).__init__()
        
#         self.output_dim = output_dim
#         self.trans = nn.Parameter(torch.Tensor(output_dim, output_dim))
#         self.trans.data.uniform_(-0.1, 0.1)

#     def compute_loss(self, y_pred, y_true, mask):
#         """
#         计算CRF损失
#         """
#         y_pred = y_pred * mask
#         y_true = y_true * mask
#         target_score = self.target_score(y_pred, y_true)
#         log_norm = self.log_norm_step(y_pred, mask)
#         log_norm = self.logsumexp(log_norm, dim=1)# 计算标量
#         return log_norm - target_score

#     def forward(self, y_pred, y_true, mask):
#         """
#         y_true: [[1, 2, 3], [2, 3, 0] ]
#         mask: [[1, 1, 1], [1, 1, 0]]
#         """
#         if y_pred.shape[0] != mask.shape[0] or y_pred.shape[1] != mask.shape[1]:
#             raise Exception("mask shape is not match to y_pred shape")
#         mask = mask.reshape((mask.shape[0], mask.shape[1], 1))
#         mask = mask.float()
#         y_true = y_true.reshape(y_pred.shape[:-1])
#         y_true = y_true.long()  
#         y_true_onehot = F.one_hot(y_true, self.output_dim) 
#         y_true_onehot = y_true_onehot.float()

#         return self.compute_loss(y_pred, y_true_onehot, mask)
    
#     def target_score(self, y_pred, y_true):
#         """
#         计算状态标签得分 + 转移标签得分
#         y_true: (batch, seq_len, out_dim)
#         y_pred: (batch, seq_len, out_dim)
#         """
#         # print(y_pred.shape)
#         # print(y_true.shape)
#         point_score = torch.einsum("bni,bni->b", y_pred, y_true)
#         trans_score = torch.einsum("bni,ij,bnj->b", y_true[:, :-1], self.trans, y_true[:, 1: ])

#         return point_score + trans_score
    
#     def log_norm_step(self, y_pred, mask):
#         """
#         计算归一化因子Z(X)
#         """
#         state = y_pred[:, 0] # 初始Z(X)
#         y_pred = y_pred[:, 1: ].contiguous()
#         mask = mask[:, 1:].contiguous()
#         batch, seq_len, out_dim = y_pred.shape
#         for t in range(seq_len):
#             cur_mask = mask[:, t]
#             state = torch.unsqueeze(state, 2) # (batch, out_dim, 1)
#             g = torch.unsqueeze(self.trans, 0) # (1, out_dim, out_dim)
#             outputs = self.logsumexp(state + g, dim=1) # batch, out_dim
#             outputs = outputs + y_pred[:, t]
#             outputs = cur_mask * outputs + (1 - cur_mask) * state.squeeze(-1)
#             state = outputs
        
#         return outputs
    
#     def logsumexp(self, x, dim=None, keepdim=False):
#         """
#         避免溢出
#         """
#         if dim is None:
#             x, dim = x.view(-1), 0
#         xm, _ = torch.max(x, dim, keepdim=True)
#         out = xm + torch.log(torch.sum(torch.exp(x - xm), dim=dim, keepdim=True))
#         return out if keepdim else out.squeeze(dim)


from typing import List, Optional

import torch
import torch.nn as nn


class CRF(nn.Module):
    """Conditional random field.

    This module implements a conditional random field [LMP01]_. The forward computation
    of this class computes the log likelihood of the given sequence of tags and
    emission score tensor. This class also has `~CRF.decode` method which finds
    the best tag sequence given an emission score tensor using `Viterbi algorithm`_.

    Args:
        num_tags: Number of tags.
        batch_first: Whether the first dimension corresponds to the size of a minibatch.

    Attributes:
        start_transitions (`~torch.nn.Parameter`): Start transition score tensor of size
            ``(num_tags,)``.
        end_transitions (`~torch.nn.Parameter`): End transition score tensor of size
            ``(num_tags,)``.
        transitions (`~torch.nn.Parameter`): Transition score tensor of size
            ``(num_tags, num_tags)``.


    .. [LMP01] Lafferty, J., McCallum, A., Pereira, F. (2001).
       "Conditional random fields: Probabilistic models for segmenting and
       labeling sequence data". *Proc. 18th International Conf. on Machine
       Learning*. Morgan Kaufmann. pp. 282–289.

    .. _Viterbi algorithm: https://en.wikipedia.org/wiki/Viterbi_algorithm
    """

    def __init__(self, num_tags: int, batch_first: bool = True) -> None:
        if num_tags <= 0:
            raise ValueError(f'invalid number of tags: {num_tags}')
        super().__init__()
        self.num_tags = num_tags
        self.batch_first = batch_first
        self.start_transitions = nn.Parameter(torch.empty(num_tags))
        self.end_transitions = nn.Parameter(torch.empty(num_tags))
        self.transitions = nn.Parameter(torch.empty(num_tags, num_tags))

        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Initialize the transition parameters.

        The parameters will be initialized randomly from a uniform distribution
        between -0.1 and 0.1.
        """
        nn.init.uniform_(self.start_transitions, -0.1, 0.1)
        nn.init.uniform_(self.end_transitions, -0.1, 0.1)
        nn.init.uniform_(self.transitions, -0.1, 0.1)


    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(num_tags={self.num_tags})'

    def forward(
            self,
            emissions: torch.Tensor,
            tags: torch.LongTensor,
            mask: Optional[torch.ByteTensor] = None,
            reduction: str = 'sum',
            ) -> torch.Tensor:
        """Compute the conditional log likelihood of a sequence of tags given emission scores.

        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length, num_tags)`` otherwise.
            tags (`~torch.LongTensor`): Sequence of tags tensor of size
                ``(seq_length, batch_size)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length)`` otherwise.
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
                if ``batch_first`` is ``False``, ``(batch_size, seq_length)`` otherwise.
            reduction: Specifies  the reduction to apply to the output:
                ``none|sum|mean|token_mean``. ``none``: no reduction will be applied.
                ``sum``: the output will be summed over batches. ``mean``: the output will be
                averaged over batches. ``token_mean``: the output will be averaged over tokens.

        Returns:
            `~torch.Tensor`: The log likelihood. This will have size ``(batch_size,)`` if
            reduction is ``none``, ``()`` otherwise.
        """
        self._validate(emissions, tags=tags, mask=mask)
        if reduction not in ('none', 'sum', 'mean', 'token_mean'):
            raise ValueError(f'invalid reduction: {reduction}')
        if mask is None:
            mask = torch.ones_like(tags, dtype=torch.uint8)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            tags = tags.transpose(0, 1)
            mask = mask.transpose(0, 1)

        # shape: (batch_size,)
        numerator = self._compute_score(emissions, tags, mask)
        # shape: (batch_size,)
        denominator = self._compute_normalizer(emissions, mask)
        # shape: (batch_size,)
        llh = numerator - denominator

        if reduction == 'none':
            return llh
        if reduction == 'sum':
            return llh.sum()
        if reduction == 'mean':
            return llh.mean()
        assert reduction == 'token_mean'
        return llh.sum() / mask.float().sum()


    def decode(self, emissions: torch.Tensor,
               mask: Optional[torch.ByteTensor] = None) -> List[List[int]]:
        """Find the most likely tag sequence using Viterbi algorithm.

        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length, num_tags)`` otherwise.
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
                if ``batch_first`` is ``False``, ``(batch_size, seq_length)`` otherwise.

        Returns:
            List of list containing the best tag sequence for each batch.
        """
        self._validate(emissions, mask=mask)
        if mask is None:
            mask = emissions.new_ones(emissions.shape[:2], dtype=torch.uint8)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            mask = mask.transpose(0, 1)

        return self._viterbi_decode(emissions, mask)


    def _validate(
            self,
            emissions: torch.Tensor,
            tags: Optional[torch.LongTensor] = None,
            mask: Optional[torch.ByteTensor] = None) -> None:
        if emissions.dim() != 3:
            raise ValueError(f'emissions must have dimension of 3, got {emissions.dim()}')
        if emissions.size(2) != self.num_tags:
            raise ValueError(
                f'expected last dimension of emissions is {self.num_tags}, '
                f'got {emissions.size(2)}')

        if tags is not None:
            if emissions.shape[:2] != tags.shape:
                raise ValueError(
                    'the first two dimensions of emissions and tags must match, '
                    f'got {tuple(emissions.shape[:2])} and {tuple(tags.shape)}')

        if mask is not None:
            if emissions.shape[:2] != mask.shape:
                raise ValueError(
                    'the first two dimensions of emissions and mask must match, '
                    f'got {tuple(emissions.shape[:2])} and {tuple(mask.shape)}')
            no_empty_seq = not self.batch_first and mask[0].all()
            no_empty_seq_bf = self.batch_first and mask[:, 0].all()
            if not no_empty_seq and not no_empty_seq_bf:
                raise ValueError('mask of the first timestep must all be on')

    def _compute_score(
            self, emissions: torch.Tensor, tags: torch.LongTensor,
            mask: torch.ByteTensor) -> torch.Tensor:
        # emissions: (seq_length, batch_size, num_tags)
        # tags: (seq_length, batch_size)
        # mask: (seq_length, batch_size)
        assert emissions.dim() == 3 and tags.dim() == 2
        assert emissions.shape[:2] == tags.shape
        assert emissions.size(2) == self.num_tags
        assert mask.shape == tags.shape
        assert mask[0].all()

        seq_length, batch_size = tags.shape
        mask = mask.float()

        # Start transition score and first emission
        # shape: (batch_size,)
        score = self.start_transitions[tags[0]]
        score += emissions[0, torch.arange(batch_size), tags[0]]

        for i in range(1, seq_length):
            # Transition score to next tag, only added if next timestep is valid (mask == 1)
            # shape: (batch_size,)
            score += self.transitions[tags[i - 1], tags[i]] * mask[i]

            # Emission score for next tag, only added if next timestep is valid (mask == 1)
            # shape: (batch_size,)
            score += emissions[i, torch.arange(batch_size), tags[i]] * mask[i]

        # End transition score
        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1
        # shape: (batch_size,)
        last_tags = tags[seq_ends, torch.arange(batch_size)]
        # shape: (batch_size,)
        score += self.end_transitions[last_tags]

        return score

    def _compute_normalizer(
            self, emissions: torch.Tensor, mask: torch.ByteTensor) -> torch.Tensor:
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        assert emissions.dim() == 3 and mask.dim() == 2
        assert emissions.shape[:2] == mask.shape
        assert emissions.size(2) == self.num_tags
        assert mask[0].all()

        seq_length = emissions.size(0)

        # Start transition score and first emission; score has size of
        # (batch_size, num_tags) where for each batch, the j-th column stores
        # the score that the first timestep has tag j
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]

        for i in range(1, seq_length):
            # Broadcast score for every possible next tag
            # shape: (batch_size, num_tags, 1)
            broadcast_score = score.unsqueeze(2)

            # Broadcast emission score for every possible current tag
            # shape: (batch_size, 1, num_tags)
            broadcast_emissions = emissions[i].unsqueeze(1)

            # Compute the score tensor of size (batch_size, num_tags, num_tags) where
            # for each sample, entry at row i and column j stores the sum of scores of all
            # possible tag sequences so far that end with transitioning from tag i to tag j
            # and emitting
            # shape: (batch_size, num_tags, num_tags)
            next_score = broadcast_score + self.transitions + broadcast_emissions

            # Sum over all possible current tags, but we're in score space, so a sum
            # becomes a log-sum-exp: for each sample, entry i stores the sum of scores of
            # all possible tag sequences so far, that end in tag i
            # shape: (batch_size, num_tags)
            next_score = torch.logsumexp(next_score, dim=1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # shape: (batch_size, num_tags)
            score = torch.where(mask[i].unsqueeze(1), next_score, score)

        # End transition score
        # shape: (batch_size, num_tags)
        score += self.end_transitions

        # Sum (log-sum-exp) over all possible tags
        # shape: (batch_size,)
        return torch.logsumexp(score, dim=1)

    def _viterbi_decode(self, emissions: torch.FloatTensor,
                        mask: torch.ByteTensor) -> List[List[int]]:
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        assert emissions.dim() == 3 and mask.dim() == 2
        assert emissions.shape[:2] == mask.shape
        assert emissions.size(2) == self.num_tags
        assert mask[0].all()

        seq_length, batch_size = mask.shape

        # Start transition and first emission
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]
        history = []

        # score is a tensor of size (batch_size, num_tags) where for every batch,
        # value at column j stores the score of the best tag sequence so far that ends
        # with tag j
        # history saves where the best tags candidate transitioned from; this is used
        # when we trace back the best tag sequence

        # Viterbi algorithm recursive case: we compute the score of the best tag sequence
        # for every possible next tag
        for i in range(1, seq_length):
            # Broadcast viterbi score for every possible next tag
            # shape: (batch_size, num_tags, 1)
            broadcast_score = score.unsqueeze(2)

            # Broadcast emission score for every possible current tag
            # shape: (batch_size, 1, num_tags)
            broadcast_emission = emissions[i].unsqueeze(1)

            # Compute the score tensor of size (batch_size, num_tags, num_tags) where
            # for each sample, entry at row i and column j stores the score of the best
            # tag sequence so far that ends with transitioning from tag i to tag j and emitting
            # shape: (batch_size, num_tags, num_tags)
            next_score = broadcast_score + self.transitions + broadcast_emission

            # Find the maximum score over all possible current tag
            # shape: (batch_size, num_tags)
            next_score, indices = next_score.max(dim=1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # and save the index that produces the next score
            # shape: (batch_size, num_tags)
            score = torch.where(mask[i].unsqueeze(1), next_score, score)
            history.append(indices)

        # End transition score
        # shape: (batch_size, num_tags)
        score += self.end_transitions

        # Now, compute the best path for each sample

        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1
        best_tags_list = []

        for idx in range(batch_size):
            # Find the tag which maximizes the score at the last timestep; this is our best tag
            # for the last timestep
            _, best_last_tag = score[idx].max(dim=0)
            best_tags = [best_last_tag.item()]

            # We trace back where the best last tag comes from, append that to our best tag
            # sequence, and trace it back again, and so on
            for hist in reversed(history[:seq_ends[idx]]):
                best_last_tag = hist[idx][best_tags[-1]]
                best_tags.append(best_last_tag.item())

            # Reverse the order because we start from the last timestep
            best_tags.reverse()
            best_tags_list.append(best_tags)

        return best_tags_list


from ...layers import CRF, SimpleDense
from ...metrics.chunk import ChunkF1, get_entities
from ...data.doc import Entity
from ...utils.make_model import PLMBaseModel
from ...utils.make_doc import convert_bio_to_entities
import torch
from typing import List

class CRFForEntityExtraction(PLMBaseModel):
    def __init__(self,
                 lr: float = 3e-5,
                 hidden_size: int = 256,
                 weight_decay: float = 0.01,
                 scheduler: str = 'linear_warmup',
                 **kwargs):
        super().__init__()

        self.plm = self.get_plm_architecture()

        self.classifier = SimpleDense(input_size=self.plm.config.hidden_size, 
                                      hidden_size=hidden_size, 
                                      output_size=len(self.hparams.id2bio))

        self.crf = CRF(len(self.hparams.id2bio))

        self.train_f1 = ChunkF1()
        self.val_f1 = ChunkF1()
        self.test_f1 = ChunkF1()
        
        
    def setup(self, stage: str):
        self.trainer.datamodule.dataset.set_transform(self.trainer.datamodule.bio_transform)


    def forward(self, input_ids, attention_mask, label_ids=None):
        x = self.plm(input_ids=input_ids, attention_mask=attention_mask).last_hidden_state
        emissions = self.classifier(x)
        mask = attention_mask.gt(0)
        if label_ids is not None :
            loss = self.crf(emissions, label_ids, mask=mask) * (-1)
            pred_ids= self.crf.decode(emissions, mask=mask)
            return loss, pred_ids
        else :
            return self.crf.decode(emissions, mask=mask)


    def shared_step(self, batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']
        label_ids = batch['tag_ids']
        #将label padding部分改为-1 
        label_ids[label_ids==-100] = -1
        loss, pred_ids = self(input_ids=input_ids, attention_mask=attention_mask,  label_ids=label_ids)
        pred_labels = []
        for ids in pred_ids:
            pred_labels.append([self.hparams.id2bio[id] for id in ids])

        true_labels = []
        for i in range(len(label_ids)):
            indice = torch.where(label_ids[i] >= 0)
            ids = label_ids[i][indice].tolist()
            true_labels.append([self.hparams.id2bio[id] for id in ids])
        return loss, pred_labels, true_labels
        

    def training_step(self, batch, batch_idx):
        loss, pred_labels, true_labels = self.shared_step(batch)
        self.train_f1(pred_labels, true_labels)
        self.log('train/f1', self.train_f1, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss':loss}


    def validation_step(self, batch, batch_idx):
        loss, pred_labels, true_labels = self.shared_step(batch)
        self.val_f1(pred_labels, true_labels)
        self.log('val/f1', self.val_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss':loss}

    
    def test_step(self, batch, batch_idx):
        loss, pred_labels, true_labels = self.shared_step(batch)
        self.test_f1(pred_labels, true_labels)
        self.log('test/f1', self.test_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss':loss}


    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': 0.0},
            {'params': self.crf.parameters(), 'lr': self.hparams.lr * 500, 'weight_decay': self.hparams.weight_decay}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler_config = self.get_scheduler_config(optimizer, self.hparams.scheduler)
        return [optimizer], [scheduler_config]


    def predict(self, text: str, device: str = 'cpu') -> List[Entity]:
        inputs = self.tokenizer(text,
                                max_length=self.hparams.plm_max_length,
                                truncation=True,
                                return_token_type_ids=False,
                                return_tensors='pt')
        inputs.to(device)
        outputs = self(**inputs)
        bio_tags = [self.hparams.id2bio[id] for id in outputs[0]]
        ents = convert_bio_to_entities(seq=bio_tags[1:-1])    #去掉cls sep 位置
        return ents


import torch
from transformers import AutoConfig, AutoTokenizer
from datasets import load_from_disk
import os
from ..utils.make_datamodule import PLMBaseDataModule


class TextPairRegressionDataModule(PLMBaseDataModule):
    '''文本对相似度数据模块
    dataset_exmaple:
        {'text_a': '左膝退变伴游离体','text_b': '单侧膝关节骨性关节病','similarity': 0}
    '''
    def __init__(self,
                dataset: str,
                plm: str,
                batch_size: int,
                transform: str):
        """
        Args:
            dataset (str): the name of the dataset.
            plm (str): the name of the plm.
            batch_size (int): the batch size in training and validation.
            transform (str): dataset transform
        """
        super().__init__()

        # 这一行代码为了保存传入的参数可以当做self.hparams的属性
        self.save_hyperparameters(logger=False)

        
    def setup(self, stage: str):
        dataset_path = os.path.join(self.hparams.dataset_dir, self.hparams.dataset)
        self.dataset = load_from_disk(dataset_path)
        plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
        self.tokenizer = AutoTokenizer.from_pretrained(plm_path)
        self.hparams['vocab'] = dict(sorted(self.tokenizer.vocab.items(), key=lambda x: x[1]))
        self.hparams['trf_config'] = AutoConfig.from_pretrained(plm_path)
        self.dataset.set_transform(transform=self.transform)

    def transform(self, examples):
        batch_text_a = examples['text_a']
        batch_text_b = examples['text_b']
        similarities = examples['similarity']
        batch = {'inputs_a': [], 'inputs_b': [], 'similarities':[]}
        for i  in range(len(batch_text_a)):
            inputs_a= self.tokenizer(batch_text_a[i], 
                                    padding='max_length', 
                                    max_length=self.hparams.max_length, 
                                    truncation=True)
            inputs_a = dict(zip(inputs_a.keys(), map(torch.tensor, inputs_a.values())))
            batch['inputs_a'].append(inputs_a)
            inputs_b = self.tokenizer(batch_text_b[i],
                                    padding='max_length', 
                                    max_length=self.hparams.max_length, 
                                    truncation=True)
            inputs_b = dict(zip(inputs_b.keys(), map(torch.tensor, inputs_b.values())))
            batch['inputs_b'].append(inputs_b)
            batch['similarities'].append(torch.tensor(similarities[i], dtype=torch.float))
        
        return batch


import math
import copy
import logging
import numpy as np
from six import iteritems

logger = logging.getLogger(__name__)


class BM25(object):
    """
    BM25模型

    Args:
        corpus (:obj:`list`):
            检索的语料
        k1 (:obj:`float`, optional, defaults to 1.5):
            取正值的调优参数，用于文档中的词项频率进行缩放控制
        b (:obj:`float`, optional, defaults to 0.75):
            0到1之间的参数，决定文档长度的缩放程度，b=1表示基于文档长度对词项权重进行完全的缩放，b=0表示归一化时不考虑文档长度因素
        epsilon (:obj:`float`, optional, defaults to 0.25):
            idf的下限值
        tokenizer (:obj:`object`, optional, defaults to None):
            分词器，用于对文档进行分词操作，默认为None，按字颗粒对文档进行分词
        is_retain_docs (:obj:`bool`, optional, defaults to True):
            是否保持原始文档

    Reference:
        [1] https://github.com/RaRe-Technologies/gensim/blob/3.8.3/gensim/summarization/bm25.py
    """  # noqa: ignore flake8"

    def __init__(
        self,
        corpus,
        k1=1.5,
        b=0.75,
        epsilon=0.25,
        tokenizer=None,
        is_retain_docs=True
    ):
        self.k1 = k1
        self.b = b
        self.epsilon = epsilon

        self.docs = None
        self.corpus_size = 0
        self.avgdl = 0
        self.doc_freqs = []
        self.idf = {}
        self.doc_len = []
        self.tokenizer = tokenizer

        if is_retain_docs:
            self.docs = copy.deepcopy(corpus)

        if tokenizer:
            self.tokenizer = tokenizer
            corpus = [self.tokenizer.tokenize(document) for document in corpus]
        else:
            corpus = [list(document) for document in corpus]

        self._initialize(corpus)

    def _initialize(self, corpus):
        """Calculates frequencies of terms in documents and in corpus. Also computes inverse document frequencies."""
        nd = {}  # word -> number of documents with word
        num_doc = 0
        for document in corpus:                        
            self.corpus_size += 1
            self.doc_len.append(len(document))
            num_doc += len(document)

            frequencies = {}
            for word in document:
                if word not in frequencies:
                    frequencies[word] = 0
                frequencies[word] += 1
            self.doc_freqs.append(frequencies)

            for word, freq in iteritems(frequencies):
                if word not in nd:
                    nd[word] = 0
                nd[word] += 1

        self.avgdl = float(num_doc) / self.corpus_size

        idf_sum = 0
        negative_idfs = []
        for word, freq in iteritems(nd):
            idf = math.log(self.corpus_size - freq + 0.5) - math.log(freq + 0.5)
            self.idf[word] = idf
            idf_sum += idf
            if idf < 0:
                negative_idfs.append(word)
        self.average_idf = float(idf_sum) / len(self.idf)

        if self.average_idf < 0:
            logger.warning(
                'Average inverse document frequency is less than zero. Your corpus of {} documents'
                ' is either too small or it does not originate from natural text. BM25 may produce'
                ' unintuitive results.'.format(self.corpus_size)
            )

        eps = self.epsilon * self.average_idf
        for word in negative_idfs:
            self.idf[word] = eps

    def get_score(self, query, index):
        score = 0.0
        doc_freqs = self.doc_freqs[index]
        numerator_constant = self.k1 + 1
        denominator_constant = self.k1 * (1 - self.b + self.b * self.doc_len[index] / self.avgdl)
        for word in query:
            if word in doc_freqs:
                df = self.doc_freqs[index][word]
                idf = self.idf[word]
                score += (idf * df * numerator_constant) / (df + denominator_constant)
        return score

    def get_scores(self, query):
        scores = [self.get_score(query, index) for index in range(self.corpus_size)]
        return scores

    def recall(self, query: str, topk=5):
        if self.tokenizer:
            query = self.tokenizer.tokenize(query)
        else: query = [s for s in query]
        scores = self.get_scores(query)
        indexs = np.argsort(scores)[::-1][:topk]

        if self.docs is None:
            return [[i, scores[i]] for i in indexs]
        else:
            return [[self.docs[i], scores[i]] for i in indexs]     


import torch.nn as nn
import torch
from ...metrics.span import SpanF1
from ...utils.make_model import PLMBaseModel
from ...layers import MultiLabelCategoricalCrossEntropy, EfficientGlobalPointer, MultiDropout
from ...tricks.adversarial_training import adversical_tricks




class GlobalPointer(PLMBaseModel):
    def __init__(self,
                 lr: float,
                 hidden_size: int=64,
                 weight_decay: float=0.01,
                 dropout: float=0.2,
                 threshold: float = 0.5,
                 adv: str =None,
                 **kwargs) : 
        super().__init__()
        ## 手动optimizer 可参考https://pytorch-lightning.readthedocs.io/en/stable/common/optimizers.html#manual-optimization
        self.automatic_optimization = False    


        self.bert = self.get_plm_architecture()
        self.classifier = EfficientGlobalPointer(input_size=self.bert.config.hidden_size, 
                                                 hidden_size=hidden_size,
                                                 output_size=len(self.hparams.label2id))

        self.dropout = MultiDropout()
        self.criterion = MultiLabelCategoricalCrossEntropy()

        self.train_metric = SpanF1()
        self.val_metric = SpanF1()
        self.test_metric = SpanF1()

    def forward(self, input_ids, token_type_ids, attention_mask=None):
        x = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        x = self.dropout(x)
        logits = self.classifier(x, mask=attention_mask)
        return logits


    def on_train_start(self) -> None:
        if self.hparams.adv :
            self.adv = adversical_tricks.get(self.hparams.adv)(self.bert)


    def shared_step(self, batch):
        span_ids = batch['label_ids']
        logits = self(input_ids=batch['input_ids'], token_type_ids=batch['token_type_ids'], attention_mask=batch['attention_mask'])
        pred = logits.ge(self.hparams.threshold).float()
        batch_size, ent_type_size = logits.shape[:2]
        y_true = span_ids.reshape(batch_size*ent_type_size, -1)
        y_pred = logits.reshape(batch_size*ent_type_size, -1)
        loss = self.criterion(y_pred, y_true)
        return loss, pred, span_ids


    def training_step(self, batch, batch_idx):
        optimizer = self.optimizers()
        optimizer.zero_grad()
        scheduler = self.lr_schedulers()
        loss, pred, true = self.shared_step(batch)
        self.manual_backward(loss)
        if self.hparams.adv == 'FGM':
            self.adv.attack()
            loss_adv,  _,  _ = self.shared_step(batch)
            self.manual_backward(loss_adv)
            self.adv.restore()
            self.log_dict({'train_loss': loss, 'adv_loss': loss_adv}, prog_bar=True)
        elif self.hparams.adv == "PGD":
            self.adv.backup_grad()
            K=3
            for t in range(K):
                self.adv.attack(is_first_attack=(t==0)) # 在embedding上添加对抗扰动, first attack时备份param.data
                if t != K-1:
                    self.zero_grad()
                else:
                    self.adv.restore_grad()
                loss_adv, _, _ = self.shared_step(batch)
                self.manual_backward(loss_adv) # 反向传播，并在正常的grad基础上，累加对抗训练的梯度
            self.adv.restore()
            self.log_dict({'train_loss': loss, 'adv_loss': loss_adv}, prog_bar=True)
        optimizer.step()
        if self.trainer.is_last_batch:
            scheduler.step()
        self.train_metric(pred, true)
        self.log('train/f1', self.train_metric, on_step=True, on_epoch=True, prog_bar=True)
        self.log_dict({'train_loss': loss}, prog_bar=True)
        return {'loss': loss}


    def validation_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.val_metric(pred, true)
        self.log('val/f1', self.val_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def test_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.test_metric(pred, true)
        self.log('test/f1', self.test_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.bert.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.bert.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [optimizer], [scheduler]



    def predict(self, text: str, device: str='cpu', threshold = None):
        if threshold is None:
            threshold = self.hparams.threshold
        inputs = self.tokenizer(
            text,
            add_special_tokens=True,
            max_length=self.hparams.max_length,
            truncation=True,
            return_tensors='pt')
        mapping = self.tokenizer(text,
                                 add_special_tokens=True,
                                 max_length=self.hparams.max_length,
                                 truncation=True,)
        inputs.to(device)
        logits = self(**inputs)
        spans_ls = torch.nonzero(logits>threshold).tolist()
        spans = []
        for span in spans_ls :
            start = span[2]
            end = span[3]
            span_text = text[start-1:end]
            spans.append([start-1, end, self.hparams.id2label[span[1]], span_text])
        return spans


import torch
import torch.nn as nn
from ...metrics.span import SpanIndexF1
from ...utils.make_model import PLMBaseModel
from ...layers import MultiLabelCategoricalCrossEntropy
from ...data.doc import Doc
from typing import List, Set, Tuple

class PointerForQuestionAnswering(PLMBaseModel):
    def __init__(self,
                 lr: float = 3e-5,
                 scheduler: str = 'linear_warmup',
                 weight_decay: float = 0.01,
                 threshold: float = 0.0,
                 **kwargs) : 
        super().__init__()

        self.plm = self.get_plm_architecture()
        self.start_classifier = nn.Linear(in_features=self.plm.config.hidden_size, out_features=1)
        self.end_classifier = nn.Linear(in_features=self.plm.config.hidden_size, out_features=1)

        self.criterion = MultiLabelCategoricalCrossEntropy()
        self.train_metric = SpanIndexF1()
        self.val_metric = SpanIndexF1()
        self.test_metric = SpanIndexF1()
        
        
    def setup(self, stage: str) -> None:
        self.trainer.datamodule.dataset.set_transform(self.trainer.datamodule.pointer_transform)


    def forward(self, input_ids, token_type_ids, attention_mask=None) -> torch.Tensor:
        x = self.plm(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        start_logits = self.start_classifier(x).squeeze(-1)
        end_logits = self.end_classifier(x).squeeze(-1)
        return start_logits, end_logits


    def shared_step(self, batch):
        mask = batch['attention_mask']
        start_tags = batch['start_tags']
        end_tags = batch['end_tags']
        start_logits, end_logits = self(input_ids=batch['input_ids'], token_type_ids=batch['token_type_ids'], attention_mask=batch['attention_mask'])
        start_loss = self.criterion(start_logits, start_tags)
        end_loss = self.criterion(end_logits, end_tags)
        loss = (start_loss + end_loss) / 2
        pred, _ = self.extract_spans(start_logits=start_logits, end_logits=end_logits, threshold=self.hparams.threshold)
        true, _ = self.extract_spans(start_logits=start_tags, end_logits=end_tags, threshold=self.hparams.threshold)
        return loss, pred, true
    
    def extract_spans(self, start_logits: torch.Tensor, end_logits: torch.Tensor, threshold: float) -> Tuple[List[Set], List[List[Tuple]]]:
        start_logits = start_logits.chunk(start_logits.shape[0])
        end_logits = end_logits.chunk(end_logits.shape[0])
        spans = []
        batch_spans = []
        batch_indices = []
        for i in range(len(start_logits)):
            starts = torch.where(start_logits[i].squeeze(0) > threshold)[0].tolist()
            ends = torch.where(end_logits[i].squeeze(0) > threshold)[0].tolist()
            length = min(len(starts), len(ends))
            starts = starts[:length]
            ends = ends[:length]
            indices = set()
            spans = []
            for i in range(len(starts)):
                start = starts[i]
                end = ends[i]
                if end >= start:
                    spans.append((start, end+1))
                    for j in range(start, end+1):
                        indices.add(j)
            batch_indices.append(indices)
            batch_spans.append(spans)
        return batch_indices, batch_spans
            
            
    def training_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch=batch)
        self.train_metric(pred, true)
        self.log('train/f1', self.train_metric, on_step=True, prog_bar=True)
        return loss


    def validation_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.val_metric(pred, true)
        self.log('val/f1', self.val_metric, on_epoch=True, prog_bar=True)


    def test_step(self, batch, batch_idx):
        loss, preds, tags = self.shared_step(batch)
        self.test_metric(preds, tags)
        self.log('test/f1', self.test_metric, on_epoch=True, prog_bar=True)


    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.start_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.start_classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.end_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.end_classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler = self.get_scheduler_config(optimizer=optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler]


    def predict(self, batch_question: List[str], batch_text: List[str], device: str='cpu'):
        inputs = self.tokenizer(batch_question,
                                batch_text,
                                max_length=512,
                                padding=True,
                                truncation=True,
                                return_tensors='pt')
        inputs.to(device)
        start_logits, end_logits = self(**inputs)
        _, batch_spans = self.extract_spans(start_logits=start_logits, end_logits=end_logits, threshold=self.hparams.threshold)
        align_batch_spans = []
        for spans in batch_spans:
            align_spans = []
            for span in spans:
                start = span[0]
                end = span[-1] - 1
                start_char_span= inputs.token_to_chars(start)
                end_char_span = inputs.token_to_chars(end)
                if start_char_span is not None and end_char_span is not None:
                    align_spans.append((start_char_span.start, end_char_span.end))
            align_batch_spans.append(align_spans)
        return align_batch_spans
    
    
    def set_annotation(self, doc: Doc, device: str = 'cpu', max_split_length: int = 450) -> Doc:
        for q, a in doc.questions.items():
            pieces = doc.split_by_sents(max_length=max_split_length)
            for piece in pieces:
                batch_text = [piece.text]
                batch_question = [q]
                spans = self.predict(batch_question=batch_question, batch_text=batch_text, device=device)[0]
                for span in spans:
                    answer_indices = piece.indices[span[0]: span[1]]
                    if len(answer_indices) > 0:
                        doc.add_answer_span(question=q, answer_indices=answer_indices)
        return doc  


from .global_pointer import GlobalPointer, EfficientGlobalPointer
from .simple_dense import SimpleDense
from .crf import CRF
from .biaffine import Biaffine, BiaffineSpanClassifier, EfficientBiaffineSpanClassifier
from .onerel import OneRelSpanClassifier